"""
Scikit learn interface - for making revrand compatible with pipelines.

Note
----
You cannot use the basis function objects in this module with the regressors if
you wish to learn their parameters, they are simply included as transformation
objects. You can still use the original basis function objects with the
regression wrappers here though.

"""

from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin

from . import slm, glm
from . import basis_functions as bf
from .btypes import Parameter, Positive


class StandardLinearModel(BaseEstimator, RegressorMixin):
    """
    Standard linear model interface class.

    This provides a scikit learn compatible interface for the slm module.

    Parameters
    ----------
        basis: Basis
            A basis object, see the basis_functions module.
        var: Parameter, optional
            observation variance initial value.
        regulariser: Parameter, optional
            weight regulariser (variance) initial value.
        tol: float, optional
            optimiser function tolerance convergence criterion.
        maxiter: int, optional
            maximum number of iterations for the optimiser.
    """

    def __init__(self, basis, var=Parameter(1., Positive()),
                 regulariser=Parameter(1., Positive()), tol=1e-6, maxiter=500):

        self.basis = basis
        self.var = var
        self.regulariser = regulariser
        self.tol = tol
        self.maxiter = maxiter

    def fit(self, X, y):
        """
        Learn the parameters and hyperparameters of a Bayesian linear
        regressor.

        Parameters
        ----------
            X: ndarray
                (N, d) array input dataset (N samples, d dimensions).
            y: ndarray
                (N,) array targets (N samples)
        """

        self.m, self.C, self.hypers, self.optvar = \
            slm.learn(X, y,
                      basis=self.basis,
                      var=self.var,
                      regulariser=self.regulariser,
                      tol=self.tol,
                      maxiter=self.maxiter,
                      )

        return self

    def predict(self, X):
        """
        Predict mean from Bayesian linear regression.

        Parameters
        ----------
            X: ndarray
                (Ns,d) array query input dataset (Ns samples, d dimensions).

        Returns
        -------
            Ey: ndarray
                The expected value of y_star for the query inputs, X_star
                of shape (N_star,).
        """

        return self._predict(X)

    def predict_proba(self, X):
        """
        Full predictive distribution from Bayesian linear regression.

        Parameters
        ----------
            X: ndarray
                (Ns,d) array query input dataset (Ns samples, d dimensions).

        Returns
        -------
            Ey: ndarray
                The expected value of y_star for the query inputs, X_star
                of shape (N_star,).
            Vf: ndarray
                The expected variance of f_star for the query inputs,
                X_star of shape (N_star,).
            Vy: ndarray
                The expected variance of y_star for the query inputs,
                X_star of shape (N_star,).
        """

        return self._predict(X, uncertainty=True)

    def _predict(self, X, uncertainty=False):

        Ey, Vf, Vy = slm.predict(X,
                                 basis=self.basis,
                                 m=self.m,
                                 C=self.C,
                                 hypers=self.hypers,
                                 var=self.optvar
                                 )

        return (Ey, Vf, Vy) if uncertainty else Ey


class GeneralisedLinearModel(BaseEstimator, RegressorMixin):
    """
    Generalised linear model interface class.

    This provides a scikit learn compatible interface for the glm module.

    Parameters
    ----------
        likelihood: Object
            A likelihood object, see the likelihoods module.
        basis: Basis
            A basis object, see the basis_functions module.
        regulariser: Parameter, optional
            weight regulariser (variance) initial value.
        postcomp: int, optional
            Number of diagonal Gaussian components to use to approximate the
            posterior distribution.
        tol: float, optional
           Optimiser relative tolerance convergence criterion.
        use_sgd: bool, optional
            If :code:`True` then use SGD (Adadelta) optimisation instead of
            L-BFGS.
        maxiter: int, optional
            Maximum number of iterations of the optimiser to run. If
            :code:`use_sgd` is :code:`True` then this is the number of complete
            passes through the data before optimization terminates (unless it
            converges first).
        batch_size: int, optional
            number of observations to use per SGD batch. Ignored if
            :code:`use_sgd=False`.
        rho: float, optional
            SGD decay rate, must be [0, 1]. Ignored if :code:`use_sgd=False`.
        epsilon: float, optional
            Jitter term for adadelta SGD. Ignored if :code:`use_sgd=False`.
        alpha: float, optional
            The percentile confidence interval (e.g. 95%) to return from
            predict_proba.
    """

    def __init__(self, likelihood, basis,
                 regulariser=Parameter(1., Positive()), postcomp=10,
                 use_sgd=True, maxiter=1000, tol=1e-7, batch_size=100, rho=0.9,
                 epsilon=1e-5, alpha=0.95):

        self.likelihood = likelihood
        self.basis = basis
        self.regulariser = regulariser
        self.postcomp = postcomp
        self.use_sgd = use_sgd
        self.maxiter = maxiter
        self.tol = tol
        self.batch_size = batch_size
        self.rho = rho
        self.epsilon = epsilon
        self.alpha = alpha

    def fit(self, X, y, likelihood_args=()):
        """

        Parameters
        ----------
            X: ndarray
                (N, d) array input dataset (N samples, d dimensions).
            y: ndarray
                (N,) array targets (N samples)
        """

        self.m, self.C, self.likelihood_hypers, self.basis_hypers = \
            glm.learn(X, y,
                      likelihood=self.likelihood,
                      basis=self.basis,
                      regulariser=self.regulariser,
                      likelihood_args=likelihood_args,
                      postcomp=self.postcomp,
                      use_sgd=self.use_sgd,
                      maxiter=self.maxiter,
                      tol=self.tol,
                      batch_size=self.batch_size,
                      rho=self.rho,
                      epsilon=self.epsilon)

        return self

    def predict(self, X, likelihood_args=()):
        """
        Predict the target expected value from the generalised linear model.

        Parameters
        ----------
        X: ndarray
            (Ns,d) array query input dataset (Ns samples, d dimensions).
        likelihood_args: sequence, optional
            sequence of arguments to pass to the likelihood function. These
            are non-learnable parameters. They can be scalars or arrays of
            length Ns.

        Returns
        -------
        Ey: ndarray
            The expected value of y_star for the query inputs, X_star of shape
            (N_star,).
        """

        Ey, _, _, _ = \
            glm.predict_moments(X,
                                likelihood=self.likelihood,
                                basis=self.basis,
                                m=self.m,
                                C=self.C,
                                likelihood_hypers=self.likelihood_hypers,
                                basis_hypers=self.basis_hypers,
                                likelihood_args=likelihood_args
                                )
        return Ey

    def predict_proba(self, X, alpha=None, likelihood_args=()):
        """
        Predicted target value and uncertainty quantiles from the generalised
        linear model.

        Parameters
        ----------
        X: ndarray
            (Ns,d) array query input dataset (Ns samples, d dimensions).
        alpha: float, optional
            The percentile confidence interval (e.g. 95%) to return from
            predict_proba. If this is None, the value in the constructor is
            used.
        likelihood_args: sequence, optional
            sequence of arguments to pass to the likelihood function. These
            are non-learnable parameters. They can be scalars or arrays of
            length Ns.

        Returns
        -------
        Ey: ndarray
            The expected value of y_star for the query inputs, X_star of shape
            (Ns,).
        ql: ndarray
            The lower end point of the interval with shape (Ns,)
        qu: ndarray
            The upper end point of the interval with shape (Ns,)
        """

        if alpha is not None:
            self.alpha = alpha

        Ey = self.predict(X, likelihood_args)
        ql, qu = glm.predict_interval(alpha=self.alpha,
                                      Xs=X,
                                      likelihood=self.likelihood,
                                      basis=self.basis,
                                      m=self.m,
                                      C=self.C,
                                      likelihood_hypers=self.likelihood_hypers,
                                      basis_hypers=self.basis_hypers,
                                      likelihood_args=likelihood_args
                                      )

        return Ey, ql, qu


class _BaseBasis(BaseEstimator, TransformerMixin, bf.Basis):

    def __init__(self):

        self.kwparams = {}

    def fit(self, X, y=None):
        """
        A do-nothing fit function that just maintains compatibility with
        scikit-learn's tranformer interface.
        """

        return self

    def transform(self, X, y=None):
        """
        Apply the basis function to X.

        Parameters
        ----------
        X: ndarray
            of shape (N, d) of observations where N is the number of samples,
            and d is the dimensionality of X.

        Returns
        -------
        ndarray:
            The basis function transform (Phi).
        """

        return self(X, **self.kwparams)

    def fit_transform(self, X, y=None, **fit_params):
        """
        Apply the basis function to X.

        Parameters
        ----------
        X: ndarray
            of shape (N, d) of observations where N is the number of samples,
            and d is the dimensionality of X.
        fit_params: sequence
            extra parameters to supply the basis function (see
            basis_functions.py).

        Returns
        -------
        ndarray:
            The basis function transform (Phi).

        Note
        ----
        This is the same as :code:`transform()` with the option of passing in
        extra parameters.
        """

        if not fit_params:
            return self.transform(X, y)
        else:
            return self(X, **fit_params)


class LinearBasis(bf.LinearBasis, _BaseBasis):
    """
    Linear basis class, basically this just prepends a columns of ones onto X

    Parameters
    ----------
    onescol: bool, optional
        If true, prepend a column of ones onto X.
    """

    def __init__(self, onescol=False):

        self.onescol = onescol
        self.kwparams = {}

        super(LinearBasis, self).__init__(onescol=onescol)


class PolynomialBasis(bf.PolynomialBasis, _BaseBasis):
    """
    Polynomial basis class, this essentially creates the concatenation,
    :math:`\\boldsymbol\Phi = [\mathbf{X}^0, \mathbf{X}^1, ..., \mathbf{X}^p]`
    where :math:`p` is the :code:`order` of the polynomial.

    Parameters
    ----------
    order: int
        the order of the polynomial to create, i.e. the last power to raise
        X to in the concatenation Phi = [X^0, X^1, ..., X^order].
    include_bias: bool, optional
        If True (default), include the bias column (column of ones which
        acts as the intercept term in a linear model)
    """

    def __init__(self, order, include_bias=True):

        self.order = order
        self.include_bias = include_bias
        self.kwparams = {}

        super(PolynomialBasis, self).__init__(order=order,
                                              include_bias=include_bias)


class RadialBasis(bf.RadialBasis, _BaseBasis):
    """
    Radial basis class.

    Parameters
    ----------
    centres: ndarray
        array of shape (Dxd) where D is the number of centres for the
        radial bases, and d is the dimensionality of X.
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).

    Note
    ----
    This will have relevance vector machine-like behaviour with uncertainty.
    """

    def __init__(self, centres, lenscale):

        self.centres = centres
        self.lenscale = lenscale
        self.kwparams = {'lenscale': lenscale}

        super(RadialBasis, self).__init__(centres=centres)


class SigmoidalBasis(bf.SigmoidalBasis, RadialBasis):
    """
    Sigmoidal Basis

    Parameters
    ----------
    centres: ndarray
        array of shape (Dxd) where D is the number of centres for the bases,
        and d is the dimensionality of X.
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """

    pass


class RandomRBF(bf.RandomRBF, _BaseBasis):
    """
    Random RBF Basis -- Approximates an RBF kernel function

    This will make a linear regression model approximate a GP with an
    (optionally ARD) RBF covariance function.

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """

    def __init__(self, nbases, Xdim, lenscale):

        self.nbases = nbases
        self.Xdim = Xdim
        self.lenscale = lenscale
        self.kwparams = {'lenscale': lenscale}
        super(RandomRBF, self).__init__(Xdim=Xdim, nbases=nbases)


class RandomLaplace(bf.RandomLaplace, RandomRBF):
    """
    Random Laplace Basis -- Approximates a Laplace kernel function

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Laplace covariance function.

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """
    pass


class RandomCauchy(bf.RandomCauchy, RandomRBF):
    """
    Random Cauchy Basis -- Approximates a Cauchy kernel function

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Laplace covariance function.

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """

    pass


class RandomMatern32(bf.RandomMatern32, RandomRBF):
    """
    Random Matern 3/2 Basis -- Approximates a Matern 3/2 kernel function

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Laplace covariance function.

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """
    pass


class RandomMatern52(bf.RandomMatern52, RandomRBF):
    """
    Random Matern 5/2 Basis -- Approximates a Matern 5/2 kernel function

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Laplace covariance function.

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """

    pass


class FastFoodRBF(bf.FastFoodRBF, RandomRBF):
    """
    Fast Food radial basis function, which is an approximation of the random
    radial basis function for a large number of bases.

    This will make a linear regression model approximate a GP with an RBF
    covariance function.

    Parameters
    ----------
    nbases: int
        a scalar for how many (unique) random bases to create approximately,
        this actually will be to the nearest larger two power.
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: scalar or ndarray
        scalar or array of shape (d,) length scales (one for each dimension
        of X).
    """

    def __init__(self, nbases, Xdim, lenscale):
        self.nbases = nbases
        self.Xdim = Xdim
        self.lenscale = lenscale
        self.kwparams = {'lenscale': lenscale}
        super(FastFoodRBF, self).__init__(Xdim=Xdim, nbases=nbases)



from collections import namedtuple

Range = namedtuple('Range', 'lower_bound upper_bound initial_val')

QueryParams = namedtuple('GPQuery', 'regressor Xs K_xxs')

RegressionParams = namedtuple('RegressionParams',
                              'X factorisation alpha kernel y')

# Folds = namedtuple('Folds', 'X y flat_y n_folds')



"""
Likelihood objects for inference within the GLM framework.

"""

from __future__ import division

import numpy as np

from scipy.stats import bernoulli, binom, poisson, norm
from scipy.special import gammaln, expit

from .btypes import Parameter, Positive
from .mathfun.special import safesoftplus, softplus, logtiny


#
# Likelihood classes
#

class Bernoulli():
    """
    Bernoulli likelihood class for (binary) classification tasks.

    A logistic transformation function is used to map the latent function from
    the GLM prior into a probability.
    """

    _params = []

    def __init__(self):

        pass

    @property
    def params(self):
        """ Get this object's Parameter types. """
        return self._params

    @params.setter
    def params(self, params):
        """ Set this object's Parameter types. """
        self._params = params

    def loglike(self, y, f):
        r"""
        Bernoulli log likelihood.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """

        ll = bernoulli.logpmf(y, expit(f))
        ll[np.isinf(ll)] = logtiny
        return ll

    def Ey(self, f):
        r""" Expected value of the Bernoulli likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[y|f]`.
        """

        return expit(f)

    def df(self, y, f):
        r"""
        Derivative of Bernoulli log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """

        return y - expit(f)

    def d2f(self, y, f):
        r"""
        Second derivative of Bernoulli log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the second derivative
            :math:`\partial^2 \log p(y|f)/ \partial f^2`
        """

        sig = expit(f)
        return (sig - 1) * sig

    def d3f(self, y, f):
        r"""
        Third derivative of Bernoulli log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the third derivative
            :math:`\partial^3 \log p(y|f)/ \partial f^3`
        """

        sig = expit(f)
        return (2 * sig - 1) * (1 - sig) * sig

    def dp(self, y, f, *args):
        r"""
        Derivative of Bernoulli log likelihood w.r.t.\  the parameters,
        :math:`\theta`.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        dp: list, float or ndarray
            the derivative
            :math:`\partial \log p(y|f, \theta)/ \partial \theta` for
            each parameter. If there is only one parameter, this is not a
            list.
        """

        return []

    def dpd2f(self, y, f, *args):
        r"""
        Partial derivative of Bernoulli log likelihood,
        :math:`\partial h(f, \theta) / \partial \theta` where
        :math:`h(f, \theta) = \partial^2 \log p(y|f, \theta)/ \partial f^2`.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        dpd2f: list or ndarray
            the derivative of the likelihood Hessian w.r.t.\
            :math:`\theta` for each parameter. If there is only one
            parameter, this is not a list.
        """

        return []

    def cdf(self, y, f):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """

        return bernoulli.cdf(y, expit(f))


class Binomial(Bernoulli):
    """
    Binomial likelihood class.

    A logistic transformation function is used to map the latent function from
    the GLM prior into a probability.
    """

    def loglike(self, y, f, n):
        r"""
        Binomial log likelihood.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """

        ll = binom.logpmf(y, n=n, p=expit(f))
        ll[np.isinf(ll)] = logtiny
        return ll

    def Ey(self, f, n):
        r""" Expected value of the Binomial likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[y|f]`.
        """

        return expit(f) * n

    def df(self, y, f, n):
        r"""
        Derivative of Binomial log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """

        return y - expit(f) * n

    def d2f(self, y, f, n):
        r"""
        Second derivative of Binomial log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        df: ndarray
            the second derivative
            :math:`\partial^2 \log p(y|f)/ \partial f^2`
        """

        sig = expit(f)
        return (sig - 1) * sig * n

    def d3f(self, y, f, n):
        r"""
        Third derivative of Binomial log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        df: ndarray
            the third derivative
            :math:`\partial^3 \log p(y|f)/ \partial f^3`
        """

        sig = expit(f)
        return (2 * sig - 1) * (1 - sig) * sig * n

    def cdf(self, y, f, n):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """

        return binom.cdf(y, n=n, p=expit(f))


class Gaussian(Bernoulli):
    """
    A univariate Gaussian likelihood for general regression tasks.

    No transformation function is needed since this is (conditionally)
    conjugate to the GLM prior.

    Parameters
    ----------
    var_init: Parameter, optional
        A scalar Parameter describing the initial point and bounds for
        an optimiser to learn the variance parameter of this object.
    """

    def __init__(self, var_init=Parameter(1., Positive())):

        self.params = var_init

    def loglike(self, y, f, var):
        r"""
        Gaussian log likelihood.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """

        return norm.logpdf(y, loc=f, scale=np.sqrt(var))

    def Ey(self, f, var):
        r""" Expected value of the Gaussian likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[y|f]`.
        """

        return f

    def df(self, y, f, var):
        r"""
        Derivative of Gaussian log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """

        return (y - f) / var

    def d2f(self, y, f, var):
        r"""
        Second derivative of Gaussian log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        df: ndarray
            the second derivative
            :math:`\partial^2 \log p(y|f)/ \partial f^2`
        """

        return - np.ones_like(f) / var

    def d3f(self, y, f, var):
        r"""
        Third derivative of Gaussian log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        df: ndarray
            the third derivative
            :math:`\partial^3 \log p(y|f)/ \partial f^3`
        """

        return np.zeros_like(f)

    def dp(self, y, f, var):
        r"""
        Derivative of Gaussian log likelihood w.r.t.\ the variance
        :math:`\sigma^2`.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        dp: float
            the derivative
            :math:`\partial \log p(y|f, \sigma^2)/ \partial \sigma^2`
            where :math:`sigma^2` is the variance.
        """

        ivar = 1. / var
        return 0.5 * (((y - f) * ivar)**2 - ivar)

    def dpd2f(self, y, f, var):
        r"""
        Partial derivative of Gaussian log likelihood,
        :math:`\partial h(f, \theta) / \partial \theta` where
        :math:`h(f, \theta) = \partial^2 \log p(y|f, \theta)/ \partial f^2`.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        dpd2f: ndarray
            the derivative of the likelihood Hessian w.r.t.\ the variance
            :math:`\sigma^2`.
        """

        return np.ones_like(f) / var**2

    def cdf(self, y, f, var):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray
            The variance of the distribution

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """

        return norm.cdf(y, loc=f, scale=np.sqrt(var))


class Poisson(Bernoulli):
    """
    A Poisson likelihood, useful for various Poisson process tasks.

    An exponential transformation function and a softplus transformation
    function have been implemented.

    Parameters
    ----------
    tranfcn: string, optional
        this may be 'exp' for an exponential transformation function,
        or 'softplus' for a softplut transformation function.
    """

    def __init__(self, tranfcn='exp'):
        """
        Construct an instance of the Poisson likelihood class.


        """

        if tranfcn == 'exp' or tranfcn == 'softplus':
            self.tranfcn = tranfcn
        else:
            raise ValueError('Invalid transformation function specified!')

    def loglike(self, y, f):
        r"""
        Poisson log likelihood.

        Parameters
        ----------
        y: ndarray
            array of integer targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """

        g = np.exp(f) if self.tranfcn == 'exp' else softplus(f)
        logg = np.log(g)
        if not np.isscalar(logg):
            logg[np.isinf(logg)] = logtiny
        else:
            logg = logtiny if np.isinf(logg) else logg
        return y * logg - g - gammaln(y + 1)

    def Ey(self, f):
        r""" Expected value of the Poisson likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[y|f]`.
        """

        return np.exp(f) if self.tranfcn == 'exp' else softplus(f)

    def df(self, y, f):
        r"""
        Derivative of Poisson log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """

        if self.tranfcn == 'exp':
            return y - np.exp(f)
        else:
            return expit(f) * (y / safesoftplus(f) - 1)

    def d2f(self, y, f):
        r"""
        Second derivative of Poisson log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the second derivative
            :math:`\partial^2 \log p(y|f)/ \partial f^2`
        """

        if self.tranfcn == 'exp':
            return - np.exp(f)
        else:
            g = safesoftplus(f)
            gp = expit(f)
            g2p = gp * (1 - gp)
            return (y - g) * g2p / g - y * (gp / g)**2

    def d3f(self, y, f):
        r"""
        Third derivative of Poisson log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the third derivative
            :math:`\partial^3 \log p(y|f)/ \partial f^3`
        """

        if self.tranfcn == 'exp':
            return self.d2f(y, f)
        else:
            g = safesoftplus(f)
            gp = expit(f)
            g2p = gp * (1 - gp)
            g3p = g2p * (1 - 2 * gp)
            return g3p * (y - g) / g + 2 * y * (gp / g)**3 \
                - 3 * y * gp * g3p / g**2

    def cdf(self, y, f):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """

        mu = np.exp(f) if self.tranfcn == 'exp' else softplus(f)
        return poisson.cdf(y, mu=mu)



"""
Implementation of Bayesian GLMs with nonparametric variational inference [1]_,
with a few modifications and tweaks.

.. [1] Gershman, S., Hoffman, M., & Blei, D. "Nonparametric variational
       inference". arXiv preprint arXiv:1206.4665 (2012).
"""


from __future__ import division

import numpy as np
import logging
from itertools import chain
from multiprocessing import Pool

from scipy.stats.distributions import gamma
from scipy.optimize import brentq, minimize

from .utils import couple, append_or_extend, atleast_list
from .mathfun.special import logsumexp
from .basis_functions import apply_grad
from .optimize import sgd, structured_sgd, structured_minimizer, logtrick_sgd,\
    logtrick_minimizer, AdaDelta
from .btypes import Bound, Positive, Parameter, get_values


# Set up logging
log = logging.getLogger(__name__)


def learn(X, y, likelihood, basis, regulariser=Parameter(1., Positive()),
          likelihood_args=(), postcomp=10, use_sgd=True, maxiter=1000,
          tol=1e-7, batch_size=100, rho=0.9, epsilon=1e-5):
    r"""
    Learn the parameters of a Bayesian generalised linear model (GLM).

    The learning algorithm uses nonparametric variational inference [1]_, and
    optionally stochastic gradients.

    Parameters
    ----------
    X: ndarray
        (N, d) array input dataset (N samples, d dimensions).
    y: ndarray
        (N,) array targets (N samples)
    likelihood: Object
        A likelihood object, see the likelihoods module.
    basis: Basis
        A basis object, see the basis_functions module.
    regulariser: Parameter, optional
        weight regulariser (variance) initial value.
    likelihood_args: sequence, optional
        sequence of arguments to pass to the likelihood function. These are
        non-learnable parameters. They can be scalars or arrays of length N.
    postcomp: int, optional
        Number of diagonal Gaussian components to use to approximate the
        posterior distribution.
    use_sgd: bool, optional
        If :code:`True` then use SGD (Adadelta) optimisation instead of L-BFGS.
    maxiter: int, optional
        Maximum number of iterations of the optimiser to run. If
        :code:`use_sgd` is :code:`True` then this is the number of mini batches
        to evaluate before termination.
    tol: float, optional
        Optimiser relative tolerance convergence criterion (only if L-BFGS
        is used as the optimiser).
    batch_size: int, optional
        number of observations to use per SGD batch. Ignored if
        :code:`use_sgd=False`.
    rho: float, optional
        SGD decay rate, must be [0, 1]. Ignored if :code:`use_sgd=False`.
    epsilon: float, optional
        Jitter term for adadelta SGD. Ignored if :code:`use_sgd=False`.

    Returns
    -------
    m: ndarray
        (D, postcomp) array of posterior weight means (D is the dimension of
        the features).
    C: ndarray
        (D, postcomp) array of posterior weight variances.
    likelihood_hypers: sequence
        learned sequence of likelihood object hyperparameters.
    basis_hypers: sequence
        learned sequence of basis object hyperparameters.

    Notes
    -----
    This approximates the posterior distribution over the weights with
    a mixture of Gaussians:

    .. math ::

        \mathbf{w} \sim \frac{1}{K} \sum^K_{k=1}
            \mathcal{N}(\mathbf{m_k}, \boldsymbol{\Psi}_k)

    where,

    .. math ::

        \boldsymbol{\Psi}_k = \text{diag}([\Psi_{k,1}, \ldots,
            \Psi_{k,D}]).

    This is so arbitrary likelihoods can be used with this algorithm, while
    still mainting flexible and tractable non-Gaussian posteriors. Additionaly
    this has the benefit that we have a reduced number of parameters to
    optimise (compared with full covariance Gaussians).

    The main differences between this implementation and the GLM in [1]_ are:
        - We use diagonal mixtures, as opposed to isotropic.
        - We do not cycle between optimising eq. 10 and 11 (objectives L1 and
          L2) in the paper. We use the full objective L2 for everything,
          including the posterior means, and we optimise all parameters
          together.

    Even though these changes make learning a little slower, and require third
    derivatives of the likelihoods, we obtain better results and we can use SGD
    straight-forwardly.

    This uses the python logging module for displaying learning status. To view
    these messages have something like,

    .. code ::

        import logging
        logging.basicConfig(level=logging.INFO)
        log = logging.getLogger(__name__)

    in your calling code.
    """

    if y.ndim != 1:
        raise ValueError("y has to be a 1-d array (single task)")
    if X.ndim != 2:
        raise ValueError("X has to be a 2-d array")

    # Shapes of things
    N, d = X.shape
    D = basis(np.atleast_2d(X[0, :]), *get_values(basis.params)).shape[1]
    K = postcomp

    # Number of hyperparameters
    nlpams = len(atleast_list(get_values(likelihood.params)))
    nbpams = len(atleast_list(get_values(basis.params)))
    tothypers = nlpams + nbpams

    # Pre-allocate here
    dm = np.zeros((D, K))
    dC = np.zeros((D, K))
    H = np.empty((D, K))

    # Make sure we get list output from likelihood parameter gradients
    lgrads = couple(lambda *a: atleast_list(likelihood.dp(*a)),
                    lambda *a: atleast_list(likelihood.dpd2f(*a)))

    # Objective function Eq. 10 from [1], and gradients of ALL params
    def L2(m, C, reg, *args):

        # Unpack data and likelihood arguments if they exist
        y, X = args[tothypers], args[tothypers + 1]
        largs = args[(tothypers + 2):] if len(likelihood_args) > 0 else ()

        # Extract parameters
        lpars, bpars, = args[:nlpams], args[nlpams:tothypers]
        lpars_largs = tuple(chain(lpars, largs))

        # Batchsize and discount factor
        M, _ = X.shape
        B = N / M

        # Basis function stuff
        Phi = basis(X, *bpars)  # M x D
        Phi2 = Phi**2
        Phi3 = Phi**3
        f = Phi.dot(m)  # M x K
        df, d2f, d3f = np.zeros((M, K)), np.zeros((M, K)), np.zeros((M, K))

        # Posterior responsability terms
        logqkk = _qmatrix(m, C)
        logqk = logsumexp(logqkk, axis=0)  # log term of Eq. 7 from [1]
        pz = np.exp(logqkk - logqk)

        # Zero starts for sums over posterior mixtures
        ll = 0
        dlpars = [np.zeros_like(p) for p in lpars]

        # Big loop though posterior mixtures for calculating stuff
        for k in range(K):

            # Common likelihood calculations
            ll += B * likelihood.loglike(y, f[:, k], *lpars_largs).sum()
            df[:, k] = B * likelihood.df(y, f[:, k], *lpars_largs)
            d2f[:, k] = B * likelihood.d2f(y, f[:, k], *lpars_largs)
            d3f[:, k] = B * likelihood.d3f(y, f[:, k], *lpars_largs)
            H[:, k] = d2f[:, k].dot(Phi2) - 1. / reg

            # Posterior mean and covariance gradients
            mkmj = m[:, k][:, np.newaxis] - m
            iCkCj = 1 / (C[:, k][:, np.newaxis] + C)
            dC[:, k] = (-((mkmj * iCkCj)**2 - 2 * iCkCj).dot(pz[:, k])
                        + H[:, k]) / (2 * K)
            dm[:, k] = (df[:, k].dot(Phi)
                        + 0.5 * C[:, k] * d3f[:, k].dot(Phi3)
                        + (iCkCj * mkmj).dot(pz[:, k])
                        - m[:, k] / reg) / K

            # Likelihood parameter gradients
            ziplgrads = zip(*lgrads(y, f[:, k], *lpars_largs))
            for l, (dp, dp2df) in enumerate(ziplgrads):
                dlpars[l] -= B * \
                    (dp.sum() + 0.5 * (C[:, k] * dp2df.dot(Phi2)).sum()) / K

        # Regulariser gradient
        dreg = 0.5 * (((m**2).sum() + C.sum()) / (reg**2 * K) - D / reg)

        # Basis function parameter gradients
        def dtheta(dPhi):
            dt = 0
            dPhiPhi = dPhi * Phi
            for k in range(K):
                dPhimk = dPhi.dot(m[:, k])
                dPhiH = d2f[:, k].dot(dPhiPhi) + \
                    0.5 * (d3f[:, k] * dPhimk).dot(Phi2)
                dt -= (df[:, k].dot(dPhimk) + (C[:, k] * dPhiH).sum()) / K
            return dt

        dbpars = apply_grad(dtheta, basis.grad(X, *bpars))

        # Objective, Eq. 10 in [1]
        L2 = 1. / K * (ll
                       - 0.5 * D * K * np.log(2 * np.pi * reg)
                       - 0.5 * (m**2).sum() / reg
                       + 0.5 * (C * H).sum()
                       - logqk.sum() + np.log(K))

        log.info("L2 = {}, reg = {}, likelihood_hypers = {}, basis_hypers = {}"
                 .format(L2, reg, lpars, bpars))

        return -L2, append_or_extend([-dm, -dC, -dreg], dlpars, dbpars)

    # Intialise m and C
    m = (np.random.randn(D, K) + np.random.randn(K)) * regulariser.value
    C = gamma.rvs(2, scale=0.5, size=(D, K))

    # Pack params
    params = append_or_extend([Parameter(m, Bound()), Parameter(C, Positive()),
                               regulariser], likelihood.params, basis.params)

    # Pack data
    likelihood_args = _reshape_likelihood_args(likelihood_args, N)
    data = (y, X) + likelihood_args

    if use_sgd is False:
        nmin = structured_minimizer(logtrick_minimizer(minimize))
        res = nmin(L2, params, tol=tol, method='L-BFGS-B', jac=True,
                   args=data, options={'maxiter': maxiter, 'maxcor': 100})
    else:
        nsgd = structured_sgd(logtrick_sgd(sgd))
        updater = AdaDelta(rho=rho, epsilon=epsilon)
        res = nsgd(L2, params, data, maxiter=maxiter, updater=updater,
                   batch_size=batch_size, eval_obj=True)

    # Unpack params
    m, C, regulariser = res.x[:3]
    likelihood_hypers = res.x[3:(3 + nlpams)]
    basis_hypers = res.x[(3 + nlpams):]

    log.info("Finished! Objective = {}, reg = {}, likelihood_hypers = {}, "
             "basis_hypers = {}, message: {}."
             .format(-res.fun, regulariser, likelihood_hypers, basis_hypers,
                     res.message))

    return m, C, likelihood_hypers, basis_hypers


def predict_moments(Xs, likelihood, basis, m, C, likelihood_hypers,
                    basis_hypers, likelihood_args=(), nsamples=100):
    r"""
    Predictive moments, in particular mean and variance, of a Bayesian GLM.

    This function uses Monte-Carlo sampling to evaluate the predictive mean and
    variance of a Bayesian GLM. The exact expressions evaluated are,

    .. math ::

        \mathbb{E}[y^*] = \int g(\mathbf{w}^T \boldsymbol\phi^{*})
            p(\mathbf{w} | \mathbf{y}, \boldsymbol\Phi) d\mathbf{w},

        \mathbb{V}[y^*] = \int \left(g(\mathbf{w}^T \boldsymbol\phi^{*})
            - \mathbb{E}[y^*]\right)^2
            p(\mathbf{w} | \mathbf{y}, \boldsymbol\Phi) d\mathbf{w},

    where :math:`g(\cdot)` is the activation (inverse link) link function used
    by the GLM, and :math:`p(\mathbf{w} | \mathbf{y}, \boldsymbol\Phi)` is the
    posterior distribution over weights (from :code:`learn`). Here are few
    concrete examples of how we can use these values,

    - Gaussian likelihood: these are just the predicted mean and variance, see
      :code:`revrand.regression.predict`
    - Bernoulli likelihood: The expected value is the probability, :math:`p(y^*
      = 1)`, i.e. the probability of class one. The variance may not be so
      useful.
    - Poisson likelihood: The expected value is similar conceptually to the
      Gaussian case, and is also a *continuous* value. The median (50%
      quantile) from :code:`predict_interval` is a discrete value. Again, the
      variance in this instance may not be so useful.

    Parameters
    ----------
    Xs: ndarray
        (Ns,d) array query input dataset (Ns samples, D dimensions).
    likelihood: Object
        A likelihood object, see the likelihoods module.
    basis: Basis
        A basis object, see the basis_functions module.
    m: ndarray
        (D,) array of regression weights (posterior).
    C: ndarray
        (D,) or (D, D) array of regression weight covariances (posterior).
    likelihood_hypers: sequence
        a sequence of parameters for the likelihood object, e.g. the
        likelihoods.Gaussian object takes a variance parameter, so this should
        be :code:`[var]`.
    basis_hypers: sequence
        A sequence of hyperparameters of the basis object.
    likelihood_args: sequence, optional
        sequence of arguments to pass to the likelihood function. These are
        non-learnable parameters. They can be scalars or arrays of length Ns.
    nsamples: int, optional
        The number of samples to draw from the posterior in order to
        approximate the predictive mean and variance.

    Returns
    -------
    Ey: ndarray
        The expected value of ys for the query inputs, Xs of shape (Ns,).
    Vy: ndarray
        The expected variance of ys (excluding likelihood noise terms) for the
        query inputs, Xs of shape (Ns,).
    Ey_min: ndarray
        The minimum sampled values of the predicted mean (same shape as Ey)
    Ey_max: ndarray
        The maximum sampled values of the predicted mean (same shape as Ey)
    """

    # Get latent function samples
    N = Xs.shape[0]
    ys = np.empty((N, nsamples))
    fsamples = sample_func(Xs, basis, m, C, basis_hypers, nsamples)

    # Push samples though likelihood expected value
    for i, f in enumerate(fsamples):
        ys[:, i] = likelihood.Ey(f, *chain(likelihood_hypers, likelihood_args))

    # Average transformed samples (MC integration)
    Ey = ys.mean(axis=1)
    Vy = ((ys - Ey[:, np.newaxis])**2).sum(axis=1) / nsamples
    Ey_min = ys.min(axis=1)
    Ey_max = ys.max(axis=1)

    return Ey, Vy, Ey_min, Ey_max


def predict_logpdf(ys, Xs, likelihood, basis, m, C, likelihood_hypers,
                   basis_hypers, likelihood_args=(), nsamples=100):
    r"""
    Predictive log-probability density function of a Bayesian GLM.

    Parameters
    ----------
    ys: ndarray
        The test observations of shape (Ns,) to evaluate under,
        :math:`\log p(y^* |\mathbf{x}^*, \mathbf{X}, y)`.
    Xs: ndarray
        (Ns,d) array query input dataset (Ns samples, D dimensions).
    likelihood: Object
        A likelihood object, see the likelihoods module.
    basis: Basis
        A basis object, see the basis_functions module.
    m: ndarray
        (D,) array of regression weights (posterior).
    C: ndarray
        (D,) or (D, D) array of regression weight covariances (posterior).
    likelihood_hypers: sequence
        a sequence of parameters for the likelihood object, e.g. the
        likelihoods.Gaussian object takes a variance parameter, so this should
        be :code:`[var]`.
    basis_hypers: sequence
        A sequence of hyperparameters of the basis object.
    likelihood_args: sequence, optional
        sequence of arguments to pass to the likelihood function. These are
        non-learnable parameters. They can be scalars or arrays of length Ns.
    nsamples: int, optional
        The number of samples to draw from the posterior in order to
        approximate the predictive mean and variance.

    Returns
    -------
    logp: ndarray
       The log probability of ys given Xs of shape (Ns,).
    logp_min: ndarray
        The minimum sampled values of the predicted log probability (same shape
        as p)
    logp_max: ndarray
        The maximum sampled values of the predicted log probability (same shape
        as p)
    """

    # Get latent function samples
    N = Xs.shape[0]
    ps = np.empty((N, nsamples))
    fsamples = sample_func(Xs, basis, m, C, basis_hypers, nsamples)

    # Push samples though likelihood pdf
    for i, f in enumerate(fsamples):
        ps[:, i] = likelihood.loglike(ys, f, *chain(likelihood_hypers,
                                                    likelihood_args))

    # Average transformed samples (MC integration)
    logp = ps.mean(axis=1)
    logp_min = ps.min(axis=1)
    logp_max = ps.max(axis=1)

    return logp, logp_min, logp_max


def predict_cdf(quantile, Xs, likelihood, basis, m, C, likelihood_hypers,
                basis_hypers, likelihood_args=(), nsamples=100):
    r"""
    Predictive cumulative density function of a Bayesian GLM.

    Parameters
    ----------
    quantile: float
        The predictive probability, :math:`p(y^* \leq \text{quantile} |
        \mathbf{x}^*, \mathbf{X}, y)`.
    Xs: ndarray
        (Ns,d) array query input dataset (Ns samples, D dimensions).
    likelihood: Object
        A likelihood object, see the likelihoods module.
    basis: Basis
        A basis object, see the basis_functions module.
    m: ndarray
        (D,) array of regression weights (posterior).
    C: ndarray
        (D,) or (D, D) array of regression weight covariances (posterior).
    likelihood_hypers: sequence
        a sequence of parameters for the likelihood object, e.g. the
        likelihoods.Gaussian object takes a variance parameter, so this
        should be :code:`[var]`.
    basis_hypers: sequence
        A sequence of hyperparameters of the basis object.
    likelihood_args: sequence, optional
        sequence of arguments to pass to the likelihood function. These are
        non-learnable parameters. They can be scalars or arrays of length Ns.
    nsamples: int, optional
        The number of samples to draw from the posterior in order to
        approximate the predictive mean and variance.

    Returns
    -------
    p: ndarray
       The probability of ys <= quantile for the query inputs, Xs of shape
       (Ns,).
    p_min: ndarray
        The minimum sampled values of the predicted probability (same shape
        as p)
    p_max: ndarray
        The maximum sampled values of the predicted probability (same shape
        as p)
    """

    # Get latent function samples
    N = Xs.shape[0]
    ps = np.empty((N, nsamples))
    fsamples = sample_func(Xs, basis, m, C, basis_hypers, nsamples)

    # Push samples though likelihood cdf
    for i, f in enumerate(fsamples):
        ps[:, i] = likelihood.cdf(quantile, f,
                                  *chain(likelihood_hypers, likelihood_args))

    # Average transformed samples (MC integration)
    p = ps.mean(axis=1)
    p_min = ps.min(axis=1)
    p_max = ps.max(axis=1)

    return p, p_min, p_max


def predict_interval(alpha, Xs, likelihood, basis, m, C, likelihood_hypers,
                     basis_hypers, likelihood_args=(), nsamples=100,
                     multiproc=True):
    """
    Predictive percentile interval (upper and lower quantiles) for a Bayesian
    GLM.

    Parameters
    ----------
    alpha: float
        The percentile confidence interval (e.g. 95%) to return.
    Xs: ndarray
        (Ns,d) array query input dataset (Ns samples, D dimensions).
    likelihood: Object
        A likelihood object, see the likelihoods module.
    basis: Basis
        A basis object, see the basis_functions module.
    m: ndarray
        (D,) array of regression weights (posterior).
    C: ndarray
        (D,) or (D, D) array of regression weight covariances (posterior).
    likelihood_hypers: sequence
        a sequence of parameters for the likelihood object, e.g. the
        likelihoods.Gaussian object takes a variance parameter, so this should
        be :code:`[var]`.
    basis_hypers: sequence
        A sequence of hyperparameters of the basis object.
    likelihood_args: sequence, optional
        sequence of arguments to pass to the likelihood function. These are
        non-learnable parameters. They can be scalars or arrays of length Ns.
    nsamples: int, optional
        The number of samples to draw from the posterior in order to
        approximate the predictive mean and variance.
    multiproc: bool, optional
        Use multiprocessing to paralellise this prediction computation.

    Returns
    -------
    ql: ndarray
        The lower end point of the interval with shape (Ns,)
    qu: ndarray
        The upper end point of the interval with shape (Ns,)
    """

    N = Xs.shape[0]

    # Generate latent function samples per observation (n in N)
    fsamples = sample_func(Xs, basis, m, C, basis_hypers, nsamples, genaxis=0)

    # Make sure likelihood_args is consistent with work
    if len(likelihood_args) > 0:
        likelihood_args = _reshape_likelihood_args(likelihood_args, N)

    # Now create work for distrbuted workers
    work = ((f[0], likelihood, likelihood_hypers, f[1:], alpha)
            for f in zip(fsamples, *likelihood_args))

    # Distribute sampling and rootfinding
    if multiproc:
        pool = Pool()
        res = pool.map(_star_rootfinding, work)
        pool.close()
        pool.join()
    else:
        res = [_rootfinding(*w) for w in work]

    # Get results of work
    ql, qu = zip(*res)
    return np.array(ql), np.array(qu)


def sample_func(Xs, basis, m, C, basis_hypers, nsamples=100, genaxis=1):
    """
    Generate samples from the posterior latent function mixtures of the GLM for
    query inputs, Xs.

    Parameters
    ----------
    Xs: ndarray
        (Ns,d) array query input dataset (Ns samples, D dimensions).
    likelihood: Object
        A likelihood object, see the likelihoods module.
    basis: Basis
        A basis object, see the basis_functions module.
    m: ndarray
        (D,) array of regression weights (posterior).
    C: ndarray
        (D,) or (D, D) array of regression weight covariances (posterior).
    basis_hypers: sequence
        A sequence of hyperparameters of the basis object.
    nsamples: int, optional
        The number of samples to draw from the posterior in order to
        approximate the predictive mean and variance.
    genaxis: int
        Axis to return samples from, i.e.
        - :code:`genaxis=1` will give you one sample at a time of f for ALL
            observations (so it will iterate over nsamples).
        - :code:`genaxis=0` will give you all samples of f for ONE
            observation at a time (so it will iterate through Xs, row by row)

    Yields
    ------
    fsamples: ndarray
        of shape (Ns,) if :code:`genaxis=1` with each call being a sample
        from the mixture of latent functions over all Ns. Or of shape
        (nsamples,) if :code:`genaxis=0`, with each call being a all samples
        for an observation, n in Ns.
    """
    D, K = m.shape

    # Generate weight samples from all mixture components
    k = np.random.randint(0, K, size=(nsamples,))
    w = m[:, k] + np.random.randn(D, nsamples) * np.sqrt(C[:, k])
    Phi = basis(Xs, *basis_hypers)  # Keep this here for speed

    # Now generate latent functions samples either colwise or rowwise
    if genaxis == 1:
        fs = (Phi.dot(ws) for ws in w.T)
    elif genaxis == 0:
        fs = (phi_n.dot(w) for phi_n in Phi)
    else:
        raise ValueError("Invalid axis to generate samples from")

    return fs


#
#  Internal Module Utilities
#


def _reshape_likelihood_args(likelihood_args, N):

    reshape_args = []
    for l in likelihood_args:
        if np.isscalar(l):
            l = l * np.ones(N)

        if (np.shape(l)[0] != N) and (len(l) != 0):
            raise ValueError("Likelihood arguments not a compatible shape!")

        reshape_args.append(l)

    return tuple(reshape_args)


# For python 2.7 compatibility instead of pool.starmap
def _star_rootfinding(args):

    return _rootfinding(*args)


def _rootfinding(fn, likelihood, likelihood_hypers, likelihood_args, alpha):

    # CDF minus percentile for quantile root finding
    predCDF = lambda q, fs, percent: \
        (likelihood.cdf(q, fs, *chain(likelihood_hypers,
                                      likelihood_args))).mean() - percent

    # Convert alpha into percentages and get (conservative) bounds for brentq
    lpercent = (1 - alpha) / 2
    upercent = 1 - lpercent
    Eyn = likelihood.Ey(fn, *chain(likelihood_hypers, likelihood_args)).mean()
    lb, ub = -1000 * max(Eyn, 1), 1000 * max(Eyn, 1)

    # Do the root finding optimisation for upper and lower quantiles
    try:
        qln = brentq(predCDF, a=lb, b=ub, args=(fn, lpercent))
    except ValueError:
        qln = np.nan

    try:
        qun = brentq(predCDF, a=lb, b=ub, args=(fn, upercent))
    except ValueError:
        qun = np.nan

    return qln, qun


def _dgausll(x, mean, dcov):
    # This is faster than calling scipy.stats.norm.logpdf

    D = len(x)
    return - 0.5 * (D * np.log(2 * np.pi) + np.log(dcov).sum()
                    + ((x - mean)**2 / dcov).sum())


def _qmatrix(m, C, k=None):

    K = m.shape[1]
    if k is None:
        logq = [[_dgausll(m[:, i], m[:, j], C[:, i] + C[:, j])
                 for i in range(K)]
                for j in range(K)]
    else:
        logq = [_dgausll(m[:, k], m[:, j], C[:, k] + C[:, j])
                for j in range(K)]

    return np.array(logq)



"""
Various Bayesian linear regression learning and prediction functions.

By using the appropriate bases, this will also yield a simple implementation of
the "A la Carte" GP [1]_.

.. [1] Yang, Z., Smola, A. J., Song, L., & Wilson, A. G. "A la Carte --
   Learning Fast Kernels". Proceedings of the Eighteenth International
   Conference on Artificial Intelligence and Statistics, pp. 1098-1106,
   2015.
"""

from __future__ import division

import logging

import numpy as np
from scipy.optimize import minimize

from .utils import append_or_extend
from .mathfun.linalg import solve_posdef
from .optimize import structured_minimizer, logtrick_minimizer
from .btypes import Parameter, Positive, get_values
from .basis_functions import apply_grad

# Set up logging
log = logging.getLogger(__name__)


def learn(X, y, basis, var=Parameter(1., Positive()),
          regulariser=Parameter(1., Positive()), tol=1e-8, maxiter=1000):
    """
    Learn the parameters and hyperparameters of a Bayesian linear regressor.

    Parameters
    ----------
        X: ndarray
            (N, d) array input dataset (N samples, d dimensions).
        y: ndarray
            (N,) array targets (N samples)
        basis: Basis
            A basis object, see the basis_functions module.
        var: Parameter, optional
            observation variance initial value.
        regulariser: Parameter, optional
            weight regulariser (variance) initial value.
        tol: float, optional
            optimiser function tolerance convergence criterion.
        maxiter: int, optional
            maximum number of iterations for the optimiser.

    Returns
    -------
        m: ndarray
            (D,) array of posterior weight means (D is the dimension of the
            features).
        C: ndarray
            (D, D) array of posterior weight variances.
        bparams: sequence
            learned sequence of basis object hyperparameters.
        float:
            learned observation variance

    Notes
    -----
        This actually optimises the evidence lower bound on log marginal
        likelihood, rather than log marginal likelihood directly. In the case
        of a full posterior convariance matrix, this bound is tight and the
        exact solution will be found (modulo local minima for the
        hyperparameters).

        This uses the python logging module for displaying learning status.
        To view these messages have something like,

        .. code ::

            import logging
            logging.basicConfig(level=logging.INFO)
            log = logging.getLogger(__name__)

        in your calling code.
    """

    if y.ndim != 1:
        raise ValueError("y has to be a 1-d array (single task)")
    if X.ndim != 2:
        raise ValueError("X has to be a 2-d array")

    N, d = X.shape
    D = basis(np.atleast_2d(X[0, :]), *get_values(basis.params)).shape[1]

    # Caches for returning optimal params
    ELBOcache = [-np.inf]
    mcache = np.zeros(D)
    Ccache = np.zeros((D, D))

    def ELBO(_var, _lambda, *_theta):

        # Get Basis
        Phi = basis(X, *_theta)                      # N x D
        PhiPhi = Phi.T.dot(Phi)

        # Posterior Parameters
        iC = np.diag(np.ones(D) / _lambda) + PhiPhi / _var
        C, logdetC = solve_posdef(iC, np.eye(D))
        m = C.dot(Phi.T.dot(y)) / _var

        # Common calcs
        TrPhiPhiC = (PhiPhi * C).sum()
        TrC = np.trace(C)
        Err = y - Phi.dot(m)
        sqErr = (Err**2).sum()
        mm = (m**2).sum()

        # Calculate ELBO
        ELBO = -0.5 * (N * np.log(2 * np.pi * _var)
                       + sqErr / _var
                       + TrPhiPhiC / _var
                       + (TrC + mm) / _lambda
                       - logdetC
                       + D * np.log(_lambda)
                       - D)

        # NOTE: In the above, TriPhiPhiC / _var = D - TrC / _lambda when we
        # analytically solve for C, but we need the trace terms for gradients
        # anyway, so we'll keep them.

        # Cache square error to compute corrected variance
        if ELBO > ELBOcache[0]:
            mcache[:] = m
            Ccache[:] = C
            ELBOcache[0] = ELBO

        log.info("ELBO = {}, var = {}, reg = {}, bparams = {}."
                 .format(ELBO, _var, _lambda, _theta))

        # Grad var
        dvar = 0.5 * (-N + (sqErr + TrPhiPhiC) / _var) / _var

        # Grad reg
        dlambda = 0.5 * ((TrC + mm) / _lambda - D) / _lambda

        # Get structured basis function gradients
        def dtheta(dPhi):
            return - (m.T.dot(Err.dot(dPhi))
                      - (dPhi.T.dot(Phi) * C).sum()) / _var

        dtheta = apply_grad(dtheta, basis.grad(X, *_theta))

        return -ELBO, append_or_extend([-dvar, -dlambda], dtheta)

    params = append_or_extend([var, regulariser], basis.params)
    nmin = structured_minimizer(logtrick_minimizer(minimize))
    res = nmin(ELBO, params, method='L-BFGS-B', jac=True, tol=tol,
               options={'maxiter': maxiter, 'maxcor': 100})
    (var, regulariser), hypers = res.x[:2], res.x[2:]

    log.info("Done! ELBO = {}, var = {}, reg = {}, hypers = {}, message = {}."
             .format(-res['fun'], var, regulariser, hypers, res.message))

    return mcache, Ccache, hypers, var


def predict(Xs, basis, m, C, hypers, var):
    """
    Predict using Bayesian linear regression.

    Parameters
    ----------
        Xs: ndarray
            (Ns,d) array query input dataset (Ns samples, d dimensions).
        basis: Basis
            A basis object, see the basis_functions module.
        m: ndarray
            (D,) array of regression weights (posterior).
        C: ndarray
            (D, D) array of regression weight covariances (posterior).
        hypers: sequence
            A sequence of hyperparameters of the basis object.
        var: float
            observation variance.

    Returns
    -------
        Ey: ndarray
            The expected value of y_star for the query inputs, X_star
            of shape (N_star,).
        Vf: ndarray
            The expected variance of f_star for the query inputs,
            X_star of shape (N_star,).
        Vy: ndarray
            The expected variance of y_star for the query inputs,
            X_star of shape (N_star,).
    """

    Phi_s = basis(Xs, *hypers)

    Ey = Phi_s.dot(m)
    Vf = (Phi_s.dot(C) * Phi_s).sum(axis=1)

    return Ey, Vf, Vf + var



""" Bound types and bounded parameter types. """

import numpy as np
from collections import namedtuple


class Bound(namedtuple('Bound', ['lower', 'upper'])):
    """
    Define bounds on a variable for the optimiser. This defaults to all
    real values allowed (i.e. no bounds).

    Parameters
    ----------
    lower : float
        The lower bound.
    upper : float
        The upper bound.

    Attributes
    ----------
    lower : float
        The lower bound.
    upper : float
        The upper bound.

    Examples
    --------
    >>> b = Bound(1e-10, upper=1e-5)
    >>> b
    Bound(lower=1e-10, upper=1e-05)
    >>> b.lower
    1e-10
    >>> b.upper
    1e-05
    >>> isinstance(b, tuple)
    True
    >>> tuple(b)
    (1e-10, 1e-05)
    >>> lower, upper = b
    >>> lower
    1e-10
    >>> upper
    1e-05
    >>> Bound(42, 10)
    Traceback (most recent call last):
        ...
    ValueError: lower bound cannot be greater than upper bound!
    """

    def __new__(cls, lower=None, upper=None):

        if lower is not None and upper is not None:
            if lower > upper:
                raise ValueError('lower bound cannot be greater than upper '
                                 'bound!')
        return super(Bound, cls).__new__(cls, lower, upper)

    def __getnewargs__(self):
        """Required for pickling!"""
        return (self.lower, self.upper)

    # TODO: Transformation details for optimiser (logistic/identity)


class Positive(namedtuple('Positive', ['lower', 'upper'])):
    """
    Define a positive only bound for the optimiser. This may induce the
    'log trick' in the optimiser (when using an appropriate decorator), which
    will ignore the 'smallest' value (but will stay above 0).

    Parameters
    ---------
    upper : float
        The largest value allowed for the optimiser to evaluate (if not using
        the log trick).

    Examples
    --------
    >>> b = Positive()
    >>> b
    Positive(lower=1e-14, upper=None)

    Since ``tuple`` (and by extension its descendents) are immutable,
    the lower bound for all instances of ``Positive`` are guaranteed to
    be positive.
    """
    def __new__(cls, upper=None):

        lower = 1e-14
        if upper is not None:
            if lower > upper:
                raise ValueError('Upper bound must be greater than {}'
                                 .format(lower))    

        return super(Positive, cls).__new__(cls, lower=lower, upper=upper)

    def __getnewargs__(self):
        """Required for pickling!"""
        return (self.upper,)


    # TODO: Transformation details for optimiser (log)


class Parameter(object):
    """
    A Parameter class that associates a value (scalar or ndarray) with a bound.

    Attributes
    ----------
    value: scalar or ndarray
        a value to associate with this parameter. This is typically used as an
        initial value for an optimizer.
    bound: Bound
        a Bound tuple that describes the valid range for all of the elements in
        value
    shape: tuple
        the shape of value, returning (1,) if value is scalar
    """

    def __init__(self, value, bounds=Bound()):

        self.value = value
        self.shape = (1,) if np.isscalar(value) else value.shape
        self.bounds = bounds


def get_values(parameters):
    """ 
    Get all of the values in a sequence of Parameter instances.

    Parameters
    ----------
    parameters: sequence or Parameter
        A sequence of Parameter objects or a single Parameter object

    Returns
    -------
    list:
        returns a list of Parameter.value calls, even for a single parameter

    Examples
    --------
    >>> params = [Parameter(np.ones(2), Bound()), Parameter(2., Positive())]
    >>> get_values(params)
    [array([ 1.,  1.]), 2.0]
    """

    if isinstance(parameters, Parameter):
        return [parameters.value]

    return [p.value for p in parameters]


def flatten_bounds(parameters):
    """
    Return flattened bounds for all of the Parameters in a sequence

    Parameters
    ----------
    parameters: sequence or Parameter
        A sequence of Parameter objects or a single Parameter object

    Returns
    -------
    list:
        a list of Bounds objects/tuples that is the length of the *total*
        number of elements in all of the Parameters values. This is a list
        even if parameters is a single parameter object

    Examples
    --------
    >>> params = [Parameter(np.ones(2), Bound()), Parameter(2., Positive())]
    >>> flatten_bounds(params)
    [Bound(lower=None, upper=None), Bound(lower=None, upper=None), Positive(lower=1e-14, upper=None)]

    """

    inflate = lambda p: [p.bounds for _ in range(np.prod(p.shape))]

    if isinstance(parameters, Parameter):
        return inflate(parameters)

    return [b for p in parameters for b in inflate(p)]
